/* wydawca - automatic release submission daemon
   Copyright (C) 2008-2013, 2017, 2019-2020 Sergey Poznyakoff

   Wydawca is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   Wydawca is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with wydawca. If not, see <http://www.gnu.org/licenses/>. */

#include <wydawca.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

typedef struct timer_slot {
    size_t refcnt;
    double real;
    double self_user;		/* user time in sec */
    double self_system;		/* system time in sec */
    double children_user;	/* user time in sec */
    double children_system;	/* system time in sec */
    struct timeval real_mark;
    struct rusage self_mark;
    struct rusage children_mark;
} WY_TIMER;

static pthread_key_t key;
static pthread_once_t key_once = PTHREAD_ONCE_INIT;

static void
timer_free(void *f)
{
    free(f);
}

static void
make_key(void)
{
    pthread_key_create(&key, timer_free);
}

#define MAX_TIMERS (MY_TIMER_SPOOL_FIRST + spool_count)

WY_TIMER *
get_thread_timer_table(void)
{
    WY_TIMER *timer_table;
    
    pthread_once(&key_once, make_key);
    if ((timer_table = pthread_getspecific(key)) == NULL) {
	timer_table = grecs_calloc(MAX_TIMERS,
				   sizeof(timer_table[0]));
	pthread_setspecific(key, timer_table);
    }
    return timer_table;
}

wydawca_timer_t
get_thread_timer(int n)
{
    if (n < 0 || n >= MAX_TIMERS)
	abort();
    return get_thread_timer_table() + n;
}

wydawca_timer_t
timer_start(int idx)
{
    wydawca_timer_t t = get_thread_timer(idx);

    if (t->refcnt++ == 0) {
	gettimeofday(&t->real_mark, NULL);
	getrusage(RUSAGE_SELF, &t->self_mark);
	getrusage(RUSAGE_CHILDREN, &t->children_mark);
    }
    return t;
}

#define DIFFTIME(now,then)\
   (((now).tv_sec - (then).tv_sec) \
    + ((double)((now).tv_usec - (then).tv_usec))/1000000)

static void
_timer_compute(wydawca_timer_t t)
{
    struct timeval real;
    struct rusage rusage;

    gettimeofday(&real, NULL);
    t->real = DIFFTIME(real, t->real_mark);
    getrusage(RUSAGE_SELF, &rusage);
    t->self_user = DIFFTIME(rusage.ru_utime, t->self_mark.ru_utime);
    t->self_system = DIFFTIME(rusage.ru_stime, t->self_mark.ru_stime);

    getrusage(RUSAGE_CHILDREN, &rusage);
    t->children_user =
	DIFFTIME(rusage.ru_utime, t->children_mark.ru_utime);
    t->children_system =
	DIFFTIME(rusage.ru_stime, t->children_mark.ru_stime);
}

static inline void
_timer_reset(wydawca_timer_t t)
{
    t->real = 0.0;
    t->self_user = 0.0;
    t->self_system = 0.0;
    t->children_user = 0.0;
    t->children_system = 0.0;
}

wydawca_timer_t
timer_get(int idx)
{
    wydawca_timer_t t = get_thread_timer(idx);
    if (t->refcnt) 
	_timer_compute(t);
    return t;
}

wydawca_timer_t
timer_stop(int idx)
{
    wydawca_timer_t t = timer_get(idx);
    if (t->refcnt)
	--t->refcnt;
    return t;
}

wydawca_timer_t
timer_reset(int idx)
{
    wydawca_timer_t t = get_thread_timer(idx);
    _timer_reset(t);
    return t;
}

double
timer_get_real(wydawca_timer_t t)
{
    return t->real;
}

double
timer_get_user(wydawca_timer_t t)
{
    return t->self_user + t->children_user;
}

double
timer_get_system(wydawca_timer_t t)
{
    return t->self_system + t->children_system;
}

char *
timer_format_time(double t)
{
    char *str = NULL;
    size_t size = 0;
    if (t < 600)
	grecs_asprintf(&str, &size, "%0.3f", t);
    else {
	long int s, m, h, d;

	s = (long int) t;
	d = s / (3600 * 24);
	s -= d * 3600 * 24;
	h = s / 3600;
	s -= h * 3600;
	m = s / 60;
	s -= m * 60;

	if (d)
	    grecs_asprintf(&str, &size, "%ld+%02ld:%02ld:%02ld", d,
			   h, m, s);
	else if (h)
	    grecs_asprintf(&str, &size, "%02ld:%02ld:%02ld", h, m, s);
	else
	    grecs_asprintf(&str, &size, "%02ld:%02ld", m, s);
    }
    if (!str)
	grecs_alloc_die();
    return str;
}

/* Cumulative statistic counters and timers */
WY_STAT_COUNTER wydawca_global_stats[WY_MAX_STAT];
WY_TIMER *wydawca_global_timer_table;
pthread_mutex_t global_stats_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t global_stats_cond = PTHREAD_COND_INITIALIZER;
static int stat_stop;
pthread_t stat_tid;

void
wydawca_stat_add(int what, size_t n)
{
    if (what >= WY_MAX_STAT) abort();
    pthread_mutex_lock(&global_stats_mutex);
    wydawca_global_stats[what] += n;
    pthread_mutex_unlock(&global_stats_mutex);
}

/* When running in cron mode, this function is called by the triplet
   processing thread when it has finished processing.  The function
   incorporates the thread-specific statistics into the cumulative
   counters. */
void
wydawca_stat_update(void)
{
    int i;
    WY_TIMER *thread_timers = get_thread_timer_table();
    pthread_mutex_lock(&global_stats_mutex);
    for (i = 0; i < MAX_TIMERS; i++) {
	wydawca_global_timer_table[i].real
	    += thread_timers[i].real;
	wydawca_global_timer_table[i].self_user
	    += thread_timers[i].self_user;
	wydawca_global_timer_table[i].self_system
	    += thread_timers[i].self_system;
	wydawca_global_timer_table[i].children_user
	    += thread_timers[i].children_user;
	wydawca_global_timer_table[i].children_system
	    += thread_timers[i].children_system;
    }
    pthread_mutex_unlock(&global_stats_mutex);
}

void
wydawca_stat_notify(int final)
{
    void *ret;
    pthread_mutex_lock(&global_stats_mutex);
    stat_stop = final;
    pthread_cond_broadcast(&global_stats_cond);
    pthread_mutex_unlock(&global_stats_mutex);
    if (final)
	pthread_join(stat_tid, &ret);
}

void
wydawca_stat_reset(void)
{
    int i;
    
    for (i = 0; i < WY_MAX_STAT; i++)
	wydawca_global_stats[i] = 0;
    for (i = 0; i < MAX_TIMERS; i++)
	_timer_reset(&wydawca_global_timer_table[i]);
}

struct micronent stat_report_schedule;

void *
wy_thr_stat(void *ptr)
{
    wy_set_cur_thread_name("WY_stat");
    pthread_mutex_lock(&global_stats_mutex);
    /* export the thread's specific stats and timer pointers to
       wydawca_global_stats and wydawca_global_stats. */
    wydawca_global_timer_table = get_thread_timer_table();
    while (!stat_stop) {
	struct timespec ts;
	micron_next_time(&stat_report_schedule, &ts);
	pthread_cond_timedwait(&global_stats_cond, &global_stats_mutex, &ts);
	wydawca_stat_log();
	notify_stat();
	wydawca_stat_reset();
    }
    pthread_mutex_unlock(&global_stats_mutex);
    return NULL;
}

void
wydawca_stat_init(void)
{
    pthread_create(&stat_tid, NULL, wy_thr_stat, NULL);
}

static char *stat_name[WY_MAX_STAT] = {
    N_("errors"),
    N_("warnings"),
    N_("bad signatures"),
    N_("access violation attempts"),
    N_("complete triplets"),
    N_("incomplete triplets"),
    N_("bad triplets"),
    N_("expired triplets"),
    N_("triplet successes"),
    N_("files uploaded"),
    N_("files archived"),
    N_("symlinks created"),
    N_("symlinks removed"),
    N_("check failures"),
};

static char *stat_kwname[WY_MAX_STAT] = {
    "stat:errors",
    "stat:warnings",
    "stat:bad_signatures",
    "stat:access_violations",
    "stat:complete_triplets",
    "stat:incomplete_triplets",
    "stat:bad_triplets",
    "stat:expired_triplets",
    "stat:triplet_success",
    "stat:uploads",
    "stat:archives",
    "stat:symlinks",
    "stat:rmsymlinks",
    "stat:check_failures"
};

int
wy_stat_mask_p(unsigned long mask)
{
    int i;

    for (i = 0; i < WY_MAX_STAT; i++)
	if (wydawca_global_stats[i] != 0 && (mask && WY_STAT_MASK(i)))
	    return 1;
    return 0;
}

int
wy_stat_expansion(char **ret, char const *name, size_t len)
{
    int i;

    for (i = 0; i < WY_MAX_STAT; i++) {
	if (strlen(stat_kwname[i]) == len
	    && memcmp(stat_kwname[i], name, len) == 0) {
	    size_t size = 0;
	    *ret = NULL;
	    if (grecs_asprintf(ret, &size, "%u", wydawca_global_stats[i]))
		return WRDSE_NOSPACE;
	    else
		return WRDSE_OK;
	}
    }
    return WRDSE_UNDEF;
}

void
wydawca_stat_log(void)
{
    int i;

    if (wy_stat_mask_p(print_stats)) {
	for (i = 0; i < WY_MAX_STAT; i++)
	    if (print_stats & WY_STAT_MASK(i))
		wy_log(LOG_INFO, "%s: %u",
		       gettext(stat_name[i]), wydawca_global_stats[i]);
    }
}

